/*
 * Copyright (C) 2002 Ben J Jackson <ben@ben.com>.  All rights reserved.
 *
 * This file is part of Instrument Trainer for Sharp Zaurus, written for
 * the Sharp Developer Challenge:  http://www.zaurus.com/dev/challenge/
 *
 * See http://www.ben.com/ for the latest version of this software.
 *
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation and appearing in the file LICENSE.GPL included in the
 * packaging of this file.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

//static char rcsid[] = "$Id: CompassInstrument.cpp,v 1.1 2002/10/06 04:17:00 bjj Rel $";

#include "CompassInstrument.h"
#include <math.h>
#include <qimage.h>
#include <qbitmap.h>

#include "Rotate.h"
#include "Embed.h"

static const QImage &
defaultCompassImage()
{
	static const QImage *img;

	if (!img)
		img = &embed_QImage("images/Compass.png");
	return *img;
}

CompassInstrument::CompassInstrument(QWidget *parent, const char *name, WFlags flags)
	:QFrame(parent, name, flags | WRepaintNoErase)
{
	setFrameShape(QFrame::Panel);
	setFrameShadow(QFrame::Raised);
	setLineWidth(1);
	setMidLineWidth(0);

	nav_radio = NULL;
	hdg = 0;

	is_slaved = FALSE;
	is_fixed = FALSE;

	pm.setOptimization(QPixmap::BestOptim);
	dirtyf = TRUE;

	connect(this, SIGNAL(headingChanged(int)), this, SLOT(dirty()));
}

void
CompassInstrument::dirty()
{
	if (testWState(WState_Visible)) {
		render();
		repaint(visibleRect(), false);
	} else
		dirtyf = TRUE;
}

void
CompassInstrument::render()
{
	QPainter p(&pm);
	if (rotatedCompass.isNull() || rotatedCompass_hdg != hdg) {
		QImage rot = RotateQImage(defaultCompassImage(), -hdg);
		rotatedCompass = rot.smoothScale(pm.width(), pm.height());
		rotatedCompass_hdg = hdg;
	}
	bitBlt(&pm, 0, 0, &rotatedCompass, 0, 0, pm.width(), pm.height());
	p.drawPixmap(0, 0, overlayPixmap());
	p.drawPixmap(0, 0, facePixmap);

	dirtyf = FALSE;
}

void
CompassInstrument::drawContents(QPainter *paint)
{
	if (dirtyf)
		render();
	paint->drawPixmap(0, 0, pm);
}

void
CompassInstrument::setHeading(int h)
{
	if (h > 360) h -= 360;
	if (h < 0) h += 360;
	if (hdg != h) {
		hdg = h;
		emit(headingChanged(h));
	}
}

void 
CompassInstrument::mouseSetHeading(const QPoint &pos, bool initial = false)
{
	static double baseangle;
	static int baseheading;

	if (is_fixed)
		return;

	int dx = pos.x() - width() / 2;
	int dy = pos.y() - height() / 2;
	double angle = atan2(dy, dx);

	if (initial) {
		baseangle = angle;
		baseheading = hdg;
	} else {
		setHeading((baseheading + (int)((baseangle - angle) * 180. / M_PI)) % 360);
	}
}

void
CompassInstrument::mouseMoveEvent(QMouseEvent *e)
{
	mouseSetHeading(e->pos());
}

void
CompassInstrument::mousePressEvent(QMouseEvent *e)
{
	if (e->button() == QMouseEvent::LeftButton)
		mouseSetHeading(e->pos(), TRUE);
}

void
CompassInstrument::setNAV(NAV *n)
{
	if (nav_radio) {
		disconnect(nav_radio, 0, this, 0);
		if (nav_radio->airplane())
			disconnect(nav_radio->airplane(), 0, this, 0);
	}
	nav_radio = n;
	if (nav_radio) {
		connect(nav_radio, SIGNAL(changed()), this, SLOT(dirty()));
		if (is_slaved && nav_radio->airplane())
			connect(nav_radio->airplane(), SIGNAL(changed()),
				this, SLOT(updateHeading()));
	}
}

// render sunken circle overlay
void
CompassInstrument::resizeEvent(QResizeEvent *ev)
{
	QFrame::resizeEvent(ev);
	QRect sz = contentsRect();
	pm.resize(sz.size());
	dirtyf = TRUE;

	QPixmap tmp(pm.size());
	QBitmap tmpmask(pm.size());
	QPainter p2(&tmp);
	QPainter p3(&tmpmask);

	tmp.fill(backgroundColor());
	const int growby = 2;
	QRect r(-growby, -growby, pm.width() + 2*growby, pm.height() + 2*growby);
	// arcs and chords are rasterized differently (I think it's just
	// an off-by-one error in qpainter) so make an arc by hand.
	QPointArray A, B;
	A.makeArc(r.x(), r.y(), r.width(), r.height(), 45*16, 180*16);
	B.makeArc(r.x(), r.y(), r.width(), r.height(), (180+45)*16, 180*16);

	p2.setPen(backgroundColor().dark());
	p2.drawPolyline(A);
	p2.setPen(backgroundColor().light());
	p2.drawPolyline(B);

	tmpmask.fill(Qt::color1);
	p3.setPen(Qt::color1);
	p3.setBrush(Qt::color0);
	// What?  No concat?!
	uint osize = A.size();
	A.resize(A.size() + B.size());
	for (uint i = 0; i < B.size(); ++i)
		A[i + osize] = B[i];
	p3.drawPolygon(A);
	
	tmp.setMask(tmpmask);

	facePixmap = tmp;
}

void
CompassInstrument::updateHeading()
{
	if (nav_radio && nav_radio->airplane())
		setHeading(nav_radio->airplane()->heading());
}
